#    Inside The Data Tunnel - My game for PyWeek 20.
#    Copyright (C) 2015  Jannik Haberbosch
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Modules
# Standard modules.
import copy

# Third party modules.
import pygame

# Own modules
from classes.Moving_Sprite import Moving_Sprite
from classes.Projectile import Projectile
from config import config
from classes.HP_Bar import HP_Bar
from classes.Text import Text

# Before we can use PyGame, we need to initialise it.
pygame.init()


class Active_Moving_Sprite( Moving_Sprite ):
	"""An active moving sprite is a game object / sprite that actively takes
	part in the game.  This is the base-class from which the malware classes
	aswell as the allies classes and the player-classes inherit.

	Both of these have the following things in common:
		- HP, hitpoints.  These can be restored or reduced.
		- They can fire projectiles.
		- Since they are an active part of the game, they can live or die.

	"""
	def __init__( self, _image, _pos_center, _direction, _speed, _turn_speed,
				  _allowed_area, _max_hitpoints, _current_hitpoints,
				  _projectile_speed, _projectile_image, _projectile_damage ):
		"""

		"""
		super( Active_Moving_Sprite, self ).__init__( _image, _pos_center,
													  _direction, _speed,
													  _turn_speed,
													  _allowed_area )

		# Hitpoints
		self.max_hitpoints = _max_hitpoints
		self.current_hitpoints = _current_hitpoints

		# Projectiles.
		self.projectile_speed = _projectile_speed
		self.projectile_image = _projectile_image
		self.projectile_damage = _projectile_damage
		self.projectiles = pygame.sprite.Group()

		# Alpha value for fading the active moving sprite out once its' HP have
		# either fallen below 0 or are 0.
		self.alpha_value = 255

		self.rotate_image()

		self.is_dead = False

		# The HP bar.
		self.hp_bar = HP_Bar(
			self.rect.width,
			5,
			( self.rect.left, self.rect.top - 7 ),
			config.GREEN,
			config.RED,
			self.max_hitpoints
		)

		self.n_collided_projectiles = 0


	def collision_with_projectiles( self, _foreign_projectiles, _enemy ):
		"""This function checks if the active moving sprite has collided with
		a foreign projectiles.  If it has another function is called that
		defines what to do.  Also, the projectile is removed on collision.

		"""
		self.n_collided_projectiles = 0
		
		collided_projectiles = pygame.sprite.spritecollide( self, _foreign_projectiles, True )
		for x in collided_projectiles:
			self.current_hitpoints -= x.damage
		
		if type( _enemy ).__name__ in ( "Player1", "Player2", "Ally" ):
			_enemy.current_ep += len( collided_projectiles ) * self.ep_on_projectile_hit

			if self.current_hitpoints <= 0:
				if type( _enemy ).__name__ == "Player1" or type ( _enemy ).__name__ == "Player2":
					_enemy.increase_score( self.kill_score )
	
		self.n_collided_projectiles = collided_projectiles
		return collided_projectiles


	def inflict_damage_from_foreign_collided_projectiles( self,
		_foreign_collided_projectiles ):
		"""This function subtracts the damage that has been inflicted to the
		active moving sprite by each specific projectile.

		"""
		for foreign_collided_sprite in _foreign_collided_projectiles:
			self.current_hitpoints -= foreign_collided_sprite.get_projectile_damage()


	def fire_projectile( self ):
		"""This function creates a projectile sprite and fires it (the sprite
		begins to move).  It adds the thus created sprite to the list of
		projectiles.

		"""
		projectile = Projectile(
			self.projectile_image,
			self.get_position_center(),
			self.projectile_speed,
			self.projectile_damage,
			self.get_direction(),
			self.allowed_area
		)
		
		self.projectiles.add( projectile )


	def restore_hp( self, _hp_amount ):
		"""This function restores some HP and heals the active moving sprite.

		"""
		self.current_hitpoints += _hp_amount
		if self.current_hitpoints > self.max_hitpoints:
			self.current_hitpoints = self.max_hitpoints


	def update_alive_status( self ):
		"""This function updates the killed status; so it may kill the sprite if
		the HP have fallen below 0 or are, in fact, 0.

		"""
		if self.current_hitpoints <= 0:
			# Remove the sprite from the screen.
			self.is_dead = True


	def update_fadeout_sprite( self ):
		"""This function fades out the the sprite step by step.  It also fades
		out projectiles.

		"""
		# We need to fade out the sprite itself and later on maybe some of its'
		# other components.
		self.alpha_value -= 10
		self.image.set_alpha( self.alpha_value )

		for p in self.projectiles:
			p.image.set_alpha( self.alpha_value )


	def is_faded_out( self ):
		"""This function checks if the sprite has been fully faded out, i.e., is
		not visible anymore on the window.

		"""
		if self.alpha_value <= 0:
			return True
		else:
			return False


	def can_move_up( self ):
		"""This function checks if the active moving sprite can still move up.
		If so, 'True' is returned, otherwise 'False'.

		"""
		if self.rect.top > config.ACTIVE_PLAYING_AREA[1]:
			return True
		else:
			return False


	def can_move_down( self ):
		"""This function checks if the active moving sprite can still move down.
		If so, 'True' is returned, otherwise 'False'.

		"""
		if self.rect.bottom < config.ACTIVE_PLAYING_AREA[3]:
			return True
		else:
			return False


	def update( self, _players ):
		"""Updates the sprite and its' related sprites which it uses.

		Parameters:
			_foreign_projectiles_groups
			list

		"""

		# First of all, we need to check whether the active moving sprite has
		# been killed or not.  If it has been killed, we will slowly fade-out
		# the sprite until it is invisible.  Then, we will kill it, and thus,
		# removing it from any sprite group it may belong to.
		if not self.is_dead:
		
			# First we need to check for any collisions with foreign projectiles.
			foreign_collided_projectiles_groups = []
			for player in _players:
				foreign_collided_projectiles_groups = self.collision_with_projectiles(
					player.projectiles, player
				)


			# Afterwards we may need to inflict damage to the active moving sprite
			# if any projectiles from all of these groups have collided with the 
			# active moving sprite.
			#print foreign_collided_projectiles_groups
			#for foreign_collided_projectiles_group in foreign_collided_projectiles_groups:
			#	self.inflict_damage_from_foreign_collided_projectiles(
			#		foreign_collided_projectiles_group
			#	)
			
			# Then we also need to check if the active moving sprite has been killed
			# or not.
			self.update_alive_status()

			

		else:
			if self.is_faded_out():
				self.kill()
				self.projectiles.empty()
			else:
				self.update_fadeout_sprite()

		self.hp_bar.update( self.current_hitpoints, ( self.rect.left, self.rect.top - 7 ) )


	def upgrade_properties( self, _max_hitpoints, _projectile_damage ):
		"""This function upgrades ship properties.

		"""
		self.max_hitpoints += _max_hitpoints
		self.projectile_damage = _projectile_damage


	def create_level_indicator( self, _current_level ):
		"""Creates a level indicator for the ship.  Only used for players and
		allies.

		"""
		lvl_indicator_props = copy.deepcopy( config.LEVEL_INDICATOR )
		lvl_indicator_props["_text"] = lvl_indicator_props["_text"] % str( _current_level )
		lvl_indicator_props["_pos_lefttop"] = ( self.rect.left, self.rect.top - 23 )
		self.level_indicator = Text( **lvl_indicator_props )
		

	def draw( self, window ):	
		window.blit( self.image, self.rect )
		for p in self.projectiles:
			p.draw( window )
			
		window.blit( self.hp_bar.image, self.hp_bar.rect )			
